import bpy
import uuid
from ..core.material_description import MaterialDescription
from ..core.utils import write_description


##---------------------------------------------------------------------------------------------------
class MaterialConverter:
    TRANSPARENT_SHADER_UUID         = "6d7d6861-0938-41db-9fc2-187e09504c96"
    TRANSPARENT_SKINNED_SHADER_UUID = "baa040d3-4e11-4bbf-b932-5c7782d2dcc4"
    OPAQUE_SHADER_UUID              = "744556b0-67b5-4329-ba4f-a04c04f92b1c"
    OPAQUE_SKINNED_SHADER_UUID      = "6abe3b22-e8a9-48ac-ab0b-b380df1da118"

    ##-----------------------------------------------------------------------------------------------
    def __init__(self):
        self.material_uuids = []
        num_materials       = len(bpy.data.materials)
        self.material_uuids = [str(uuid.uuid4()) for _ in range(num_materials)]

    ##-----------------------------------------------------------------------------------------------
    def convert_materials(self, output_folder_path):
        materials = bpy.data.materials
        for i, material in enumerate(materials):
            # Skip if it's a Grease Pencil material
            if self.is_grease_pencil(material):
                print(f"Skipping Grease Pencil material: {material.name}")
                continue
            material_uuid = self.material_uuids[i]
            material_desc = self.create_material_description(material, material_uuid)
            write_description(material_desc.get_dict(), f"desc.material.{material_uuid}", output_folder_path)

    ##-----------------------------------------------------------------------------------------------
    def create_material_description(self, material, material_uuid):
        if self.is_transparent(material):
            shader_ref = self.TRANSPARENT_SHADER_UUID
            skinned_shader_ref = self.TRANSPARENT_SKINNED_SHADER_UUID
        else:
            shader_ref = self.OPAQUE_SHADER_UUID
            skinned_shader_ref = self.OPAQUE_SKINNED_SHADER_UUID

        mat_desc_obj = MaterialDescription(
            uuid = material_uuid,
            name = material.name,
            data_json = {},
            is_double_sided = not material.use_backface_culling,
            shader_ref = shader_ref,
            skinned_shader_ref = skinned_shader_ref
        )
        self.fill_material_description(material, mat_desc_obj)
        return mat_desc_obj
    
   ##-----------------------------------------------------------------------------------------------
    def is_grease_pencil(self,material):
        # Check if the material is used by a Grease Pencil object
        return material.grease_pencil is not None
    
    ##-----------------------------------------------------------------------------------------------
    def is_transparent(self, material):
        # Check for transparency (alpha value or alpha blending in the material settings)
        if hasattr(material, 'blend_method') and material.blend_method in {'BLEND', 'CLIP'}:
            return True
        if hasattr(material, 'alpha') and material.alpha < 1.0:
            return True
        return False
    
    ##-----------------------------------------------------------------------------------------------
    def fill_material_description(self, material, mat_desc_obj):
        # Albedo (Diffuse Color)
        mat_desc_obj.data_json["albedo"] = list(material.diffuse_color[:3])
        # Roughness and Metallic
        roughness = material.roughness if hasattr(material, 'roughness') else 1.0
        metallic = 0.0

        # Emission (from emission node if available)
        mat_desc_obj.data_json["emission"] = [0.0, 0.0, 0.0]
        if material.node_tree:
            for node in material.node_tree.nodes:
                if node.type == 'EMISSION':
                    emission_color = list(node.inputs['Color'].default_value[:3])
                    mat_desc_obj.data_json["emission"] = emission_color
                if node.type == 'BSDF_PRINCIPLED':
                    roughness  = node.inputs['Roughness'].default_value
                    metallic   = node.inputs['Metallic'].default_value 
                         
            # Check if alpha/transparency is being used in the Principled BSDF shader  
            if self.is_transparent(material):
                opacity = material.alpha
                mat_desc_obj.data_json["opacity"]   = opacity
            mat_desc_obj.data_json["roughness"] = roughness
            mat_desc_obj.data_json["metallic"]  = metallic
        
    ##-----------------------------------------------------------------------------------------------
    def does_material_have_texture(self, material, texture_type):
        if not material.node_tree:
            return False

        for node in material.node_tree.nodes:
            if node.type == 'TEX_IMAGE' and texture_type in node.label.lower():
                return True
        return False

    ##-----------------------------------------------------------------------------------------------       
    def get_material_uuid(self, material):
        for i, mat in enumerate(bpy.data.materials):
            if mat == material:
                return self.material_uuids[i]
        return "00000000-0000-0000-0000-000000000000"
