
from ..core.entity import Entity
from ..core.scene_builder import SceneBuilder
from ..core import components
from ..core.utils import write_description

import uuid
from collections import deque 
import bpy


##---------------------------------------------------------------------------------------------------
class SceneConverter():
    ##-----------------------------------------------------------------------------------------------
    def __init__(self, mesh_converter, material_converter, scene_name):        
        self.scene_builder      = SceneBuilder(scene_name)
        self.mesh_converter     = mesh_converter
        self.material_converter = material_converter

    ##-----------------------------------------------------------------------------------------------  
    def convert(self, output_folder_path):
        self.scene_builder.description.triangle_count  = self.mesh_converter.get_total_triangle_count()
        self.convert_entities()
        temp_filename       = f"desc.scene.{self.scene_builder.description.uuid}"
        write_description(self.scene_builder.description.to_dict(), temp_filename, output_folder_path)
 
    ##-----------------------------------------------------------------------------------------------
    def convert_entities(self):
        root_objects = bpy.context.scene.objects
        for obj in root_objects:
            if obj.type not in {'LIGHT', 'CAMERA'}:
                self.convert_branch(obj)
            print("Scene export finished.")

    ##-----------------------------------------------------------------------------------------------
    def convert_branch(self, root_object):
        self.construct_scene_graph(root_object)
        
        if not self.scene_builder.description.entities:
            print("No entities found, skipping...")
            return
        
        bb_min = list(root_object.bound_box[0])
        bb_max = list(root_object.bound_box[6])

        # Enlarge the scene bounding box to include the current object's bounding box
        self.scene_builder.description.aabb.enlarge(bb_min, bb_max)
            
    ##----------------------------------------------------------------------------------------------- 
    def construct_scene_graph(self, root_object):
        open_nodes = deque([(root_object, -1)])

        while open_nodes:
            current_object, parent_entity_index = open_nodes.pop()
            node_entity_index = self.generate_entity(current_object, parent_entity_index)
            if current_object.type == 'MESH':
                self.compute_mesh_entities(current_object, node_entity_index, False)

            for child_object in current_object.children:
                open_nodes.append((child_object, node_entity_index))

    ##-----------------------------------------------------------------------------------------------
    def generate_entity(self, obj, parent_entity_index):
        euid           = components.EUID(str(uuid.uuid4()))
        is_root        = parent_entity_index == -1
        local_aabb     = self.get_local_aabb(obj)
        lineage        = components.Lineage(("00000000-0000-0000-0000-000000000000"))
        
        if local_aabb:
            entity_local_aabb = components.LocalAABB(local_aabb['min'], local_aabb['max'])

        if not is_root:
            parent_entity   = self.scene_builder.get_entity(parent_entity_index)
            parent_euid     = parent_entity.euid.value
            lineage         = components.Lineage(parent_euid)

        orientation,position,scale = self.get_local_transform(obj)
        new_entity = Entity(
            debug_name      = components.DebugName(obj.name), 
            euid            = euid, 
            local_transform = components.LocalTransform(orientation, position, scale),
            local_aabb      = entity_local_aabb, 
            lineage         = lineage,
        )

        entity_index = self.scene_builder.add_entity(new_entity)
        return entity_index

    ##-----------------------------------------------------------------------------------------------   
    def compute_mesh_entities(self, obj, parent_entity_index, build_local_scene_graph):
        mesh_lineage = components.Lineage(self.scene_builder.get_entity(parent_entity_index).euid.value)
        submesh_index     = self.get_submesh_index(obj)
        mesh_uuid         = self.mesh_converter.mesh_uuid 
        material_uuid     = self.get_material_uuid(obj) 
        mesh_entity       = self.generate_mesh_entity(obj, mesh_uuid, submesh_index, mesh_lineage, material_uuid)
        self.scene_builder.add_entity(mesh_entity)

    ##-----------------------------------------------------------------------------------------------
    def generate_mesh_entity(self, obj, mesh_uuid, submesh_index, mesh_lineage, material_uuid):
        mesh_name   = obj.name if obj.name and not any(char in obj.name for char in r'\/:*?"<>|') else f"Mesh{submesh_index + 1}"
        local_aabb  = self.get_local_aabb(obj)
        orientation, position, scale = self.get_local_transform(obj)
        identity_transform = components.LocalTransform(orientation, position, scale)
        
        new_entity = Entity(
            components.DebugName(mesh_name), 
            components.EUID(str(uuid.uuid4())), 
            identity_transform,
            components.LocalAABB(local_aabb['min'], local_aabb['max']),
            mesh_lineage
        )
        new_entity.material_ref    = components.MaterialRef(material_uuid)
        new_entity.mesh_ref        = components.MeshRef(mesh_uuid,submesh_index)

        return new_entity
    
    ##-----------------------------------------------------------------------------------------------   
    def get_local_transform(self, obj):
        loc, rot, scale     = obj.matrix_world.decompose()
        orientation         = [rot.x, rot.y, rot.z, rot.w]
        position            = [loc.x, loc.y, loc.z]
        scale               = [scale.x, scale.y, scale.z]
        
        return orientation,position,scale

    ##-----------------------------------------------------------------------------------------------
    def get_material_uuid(self, obj):
        # Get the first material of the object
        if obj.data.materials:
            material = obj.data.materials[0]
            material_uuid = self.material_converter.get_material_uuid(material)  # Retrieve UUID from MaterialConverter
            return material_uuid
        return "00000000-0000-0000-0000-000000000000"
      
    ##-----------------------------------------------------------------------------------------------
    def get_local_aabb(self, obj):
        local_bbox = [corner for corner in obj.bound_box]
        min_coords = list(local_bbox[0])
        max_coords = list(local_bbox[0])

        for corner in local_bbox:
            for i in range(3):
                min_coords[i] = min(min_coords[i], corner[i])
                max_coords[i] = max(max_coords[i], corner[i])
        
        return {'min': min_coords, 'max': max_coords}

    ##-----------------------------------------------------------------------------------------------
    def get_submesh_index(self, obj):
        mesh_data = obj.data  # obj.data to match the keys in mesh_map
        if mesh_data in self.mesh_converter.blender_mesh_to_submesh_index:
            return self.mesh_converter.blender_mesh_to_submesh_index[mesh_data]
        else:
            print(f"Warning: {obj.name} not found in mesh_map.")
            return -1
    